import 'package:data_db/data_db.dart';
import 'dart:io';
import 'package:file_utils/file_utils.dart';

class _Column {
  late String name;
  late int typeId;
  dynamic? defaulValue;
  late bool isNullable;
  int? maxLength;
  int? precision;
  int? scale;
  late bool isPk;
  late String fieldType;
  late String primitiveType;

  _Column(this.name, this.typeId, this.isNullable, this.isPk,
      [this.defaulValue, this.maxLength, this.precision, this.scale]) {
    fieldType = getFieldType();
    primitiveType = getPrimitiveType();
  }
  _Column.fromRecord(Map<String, DataField> record) {
    record.getValue('column_name').map((a) => a.tryGetValue((value) {
          if (value is String) {
            name = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('type_id').map((a) => a.tryGetValue((value) {
          if (value is int) {
            typeId = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('column_default').map((a) => a.tryGetValue((value) {
          defaulValue = value;
          return true;
        }));
    record.getValue('is_nullable').map((a) => a.tryGetValue((value) {
          if (value is bool) {
            isNullable = value;
            return true;
          } else {
            return false;
          }
        }));
    record
        .getValue('character_maximum_length')
        .map((a) => a.tryGetValue((value) {
              if (value is int) {
                maxLength = value;
                return true;
              } else {
                return false;
              }
            }));
    record.getValue('numeric_precision').map((a) => a.tryGetValue((value) {
          if (value is int) {
            precision = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('numeric_scale').map((a) => a.tryGetValue((value) {
          if (value is int) {
            scale = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('is_pk').map((a) => a.tryGetValue((value) {
          if (value is bool) {
            isPk = value;
            return true;
          } else {
            return false;
          }
        }));
    fieldType = getFieldType();
    primitiveType = getPrimitiveType();
  }

  String getFieldType() {
    if ((typeId == 16)) {
      if (defaulValue != null) {
        return 'BooleanField_WithDefault_PG';
      } else {
        return 'BooleanField_PG';
      }
    } else if ((typeId == 1266) || (typeId == 1083)) {
      if (defaulValue != null) {
        return 'TimeField_WithDefault_PG';
      } else {
        return 'TimeField_PG';
      }
    } else if ((typeId == 1184) || (typeId == 1114)) {
      if (defaulValue != null) {
        return 'DateTimeField_WithDefault_PG';
      } else {
        return 'DateTimeField_PG';
      }
    } else if ((typeId == 1082)) {
      if (defaulValue != null) {
        return 'DateField_WithDefault_PG';
      } else {
        return 'DateField_PG';
      }
    } else if ((typeId == 20) ||
        (typeId == 21) ||
        (typeId == 23) ||
        (typeId == 26)) {
      if (defaulValue != null) {
        return 'IntegerField_WithDefault_PG';
      } else {
        return 'IntegerField_PG';
      }
    } else if ((typeId == 700) ||
        (typeId == 701) ||
        (typeId == 790) ||
        (typeId == 1700)) {
      if (defaulValue != null) {
        return 'NumericField_WithDefault_PG';
      } else {
        return 'NumericField_PG';
      }
    } else if ((typeId == 25) ||
        (typeId == 19) ||
        (typeId == 18) ||
        (typeId == 1043)) {
      if (defaulValue != null) {
        return 'StringField_WithDefault_PG';
      } else {
        return 'StringField_PG';
      }
    } else {
      return 'DynamicField';
    }
  }

  String getCreaateField() {
    var createFieldContent =
        '_' "$name = $fieldType(this, '$name', ${!isNullable}, $isPk";
    if (fieldType.contains('StringField')) {
      if ((maxLength != null)) {
        createFieldContent += ', $maxLength';
      } else {
        createFieldContent += ', -1';
      }
    }
    if (defaulValue != null) {
      if (fieldType.contains('String')) {
        createFieldContent += ", '$defaulValue'";
      } else if (fieldType == 'TimeField_WithDefault_PG') {
        createFieldContent +=
            ", DateFormat('HH:mm:ss').parse(${(defaulValue as String).substring(0, 9)}')";
      } else if (fieldType == 'DateTimeField_WithDefault_PG') {
        if ((defaulValue as String)[21] == ':') {
          createFieldContent +=
              ", DateFormat('yyyy-MM-dd HH:mm:ss.zzz').parse(${(defaulValue as String).substring(0, 20)}')";
        } else {
          createFieldContent +=
              ", DateFormat('yyyy-MM-dd HH:mm:ss.zzz').parse(${(defaulValue as String).substring(0, 24)}')";
        }
      } else if (fieldType == 'DateField_WithDefault_PG') {
        createFieldContent +=
            ", DateFormat('yyyy-MM-dd').parse(${(defaulValue as String).substring(0, 11)}')";
      } else {
        createFieldContent += ', $defaulValue';
      }
    }
    createFieldContent += ');';
    return createFieldContent;
  }

  String getPrimitiveType() {
    if (fieldType.contains('Integer')) {
      return 'int';
    } else if (fieldType.contains('String')) {
      return 'String';
    } else if (fieldType.contains('Numeric')) {
      return 'double';
    } else if (fieldType.contains('Boolean')) {
      return 'bool';
    } else if (fieldType.contains('Date')) {
      return 'DateTime';
    } else if (fieldType.contains('Time')) {
      return 'DateTime';
    } else {
      return 'dynamic';
    }
  }
}

extension Snake on String {
  String add(String complement) {
    return '$this\n$complement';
  }

  String capitalize() {
    return '${this[0].toUpperCase()}${substring(1)}';
  }
}

List<_Column> deserializeDataset(Dataset dataset) {
  var list = <_Column>[];
  dataset.forEach((record) {
    list.add(_Column.fromRecord(record));
  });

  return list;
}

void gernerateTableORM(Dataset dataset, bool schemaInName,
    {String? path}) async {
  var columns = deserializeDataset(dataset);
  var content = getImports(columns)
      .add(getDataFields(columns))
      .add(getOrm(dataset, columns, schemaInName));
  path ??= 'orm';
  writeInFile(path, getFileName(dataset, schemaInName), content);
}

String getFileName(Dataset dataset, bool schemaInName) {
  var schema = '';
  var name = '';
  if (dataset.tryGetValue('table_schema', (table_schema) {
    return dataset.tryGetValue('table_name', (table_name) {
      schema = (table_schema as String).toLowerCase();
      name = (table_name as String).toLowerCase();
      return true;
    });
  })) {
    if (schemaInName) {
      return 'orm_${schema}_$name.dart';
    } else {
      return 'orm_$name.dart';
    }
  } else {
    return '';
  }
}

void writeInFile(String filePath, String fileName, String contents) {
  FileUtils.mkdir([filePath]);
  File(filePath + '/' + fileName).writeAsString(contents);
}

String getImports(List<_Column> columns) {
  var content = "import 'package:static_postgres_orm/static_postgres_orm.dart';"
      .add("import 'package:dartz/dartz.dart';");
  if (columns.any((element) => (element.fieldType.contains('Date') ||
      element.fieldType.contains('Time')))) {
    content = content.add("import 'package:intl/intl.dart';");
  }

  return content;
}

String getDataFields(List<_Column> columns) {
  return ''
      .add('class _DataFields extends GenericDataFields {')
      .add(getDataFieldsDeclaration(columns))
      .add(getDataFieldCreate())
      .add(getDatafieldGetterAndSetters())
      .add(getCreateFields(columns))
      .add(getDataFieldOthersFunctions())
      .add(getAddFields(columns))
      .add('}');
}

String getDataFieldsDeclaration(List<_Column> columns) {
  var result = '';
  columns.forEach((column) {
    var s = '  late final ${column.fieldType} ' '_' '${column.name};';
    result = result.add(s);
  });
  return result;
}

String getDataFieldCreate() {
  return ''
      .add('  _DataFields() : super() {')
      .add('    fields = <GenericField>[];')
      .add('    _createFields();')
      .add('    _addFields();')
      .add('  }');
}

String getDatafieldGetterAndSetters() {
  return ''
      .add('  _DataFields get _backup => getBackup(this) as _DataFields;')
      .add('  set _backup(_DataFields v) => setBackup(this, v);');
}

String getCreateFields(List<_Column> columns) {
  var content = ''.add('  void _createFields() {');
  columns.forEach((element) {
    content = content.add('    ' + element.getCreaateField());
  });

  return content.add('  }');
}

String getDataFieldOthersFunctions() {
  return ''
      .add('  GenericField getCooField(_DataFields origin, String field) {')
      .add('    return origin.fields')
      .add('      .firstWhere((element) => getFieldName(element) == field);')
      .add('  }')
      .add('')
      .add('  void _cloneField(_DataFields origin) {')
      .add('    fields.forEach((element) {')
      .add('      element.copy(getCooField(origin, getFieldName(element)));')
      .add('    });')
      .add('  }')
      .add('')
      .add('  @override')
      .add('  void backup() {')
      .add('    _backup = _DataFields();')
      .add('    _backup._cloneField(this);')
      .add('  }')
      .add('')
      .add('  @override')
      .add('  void restore() {')
      .add('    if (assigned) {')
      .add('      _cloneField(_backup);')
      .add('      finalize();')
      .add('    }')
      .add('  }');
}

String getAddFields(List<_Column> columns) {
  var content = ''.add('  void _addFields() {');
  columns.forEach((element) {
    content = content.add('    fields.add(' '_' '${element.name});');
  });

  return content.add('  }');
}

String getOrm(Dataset dataset, List<_Column> columns, bool schemaInName) {
  var schema = '';
  var name = '';
  if (dataset.tryGetValue('table_schema', (table_schema) {
    return dataset.tryGetValue('table_name', (table_name) {
      schema = (table_schema as String).toLowerCase();
      name = (table_name as String).toLowerCase();
      return true;
    });
  })) {
    var className = '';
    if (schemaInName) {
      className = '${schema.capitalize()}_${name.capitalize()}_ORM';
    } else {
      className = '${name.capitalize()}_ORM';
    }

    return ''
        .add('class $className extends PostgressORM{')
        .add('  static final Event _addRecord = Event();')
        .add(
            '  _DataFields get _dataFields => getRecords(this)[rowIndex] as _DataFields;')
        .add('')
        .add(getFieldGetters(columns))
        .add('')
        .add('  $className(Postgres_SqlConnection sqlConnection)')
        .add("      : super(sqlConnection, '$schema', '$name', _addRecord) {")
        .add('    _addRecord.action = _executeAddRecord;')
        .add('    getRecords(this).add(_DataFields());')
        .add('  }')
        .add('')
        .add('  void _executeAddRecord() {')
        .add('    getRecords(this).add(_DataFields());')
        .add('  }')
        .add(getMaterialize(columns))
        .add(getDeleteRecord(columns))
        .add('}');
  } else {
    return '';
  }
}

String getFieldGetters(List<_Column> columns) {
  var content = '';
  columns.forEach((column) {
    content = content.add(
        '  ${column.fieldType} get ${column.name} => _dataFields._${column.name};');
  });
  return content;
}

String getMaterialize(List<_Column> columns) {
  var content = ''.add(
      '  Future<Either<ErrorSqlResult, SelectSuccesSqlResult>> materialize(');
  var pk = columns.where((element) => element.isPk);
  var params = '';
  pk.forEach((element) {
    params += ', ${element.primitiveType} ${element.name}';
  });
  content = content
      .add('      ' + params.replaceFirst(', ', '') + ') async {')
      .add('    return getMaterialize(')
      .add('        this,')
      .add('        (<GenericField>[])');

  pk.forEach((element) {
    content = content
        .add(
            '          ..add(${element.fieldType}.clone(_dataFields._${element.name})')
        .add('            ..setValue(${element.name}))');
  });

  content += ');';

  return content.add('  }');
}

String getDeleteRecord(List<_Column> columns) {
  var content = ''.add(
      '  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> deleteRecord(');
  var pk = columns.where((element) => element.isPk);
  var params = '';
  pk.forEach((element) {
    params += ', ${element.primitiveType} ${element.name}';
  });
  content = content
      .add('    ' + params.replaceFirst(', ', '') + ') async{')
      .add('    return getDeleteRecord(')
      .add('         this,')
      .add('         (<GenericField>[])');

  pk.forEach((element) {
    content = content
        .add(
            '          ..add(${element.fieldType}.clone(_dataFields._${element.name})')
        .add('            ..setValue(${element.name}))');
  });

  content += ');';

  return content.add('  }');
}
